// callbacks.c
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
 
#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "fileio.h"
#include <dirent.h>
#include "v4lif.h"
#include <math.h>
#include <string.h>
#include <time.h>
#include "ccvt.h" 

static guchar rgb_buf[3*640*480];	// main 640x480 live display image buffer
static guchar ds_buf[3*320*240];	// temp for downsized live image
static guchar ds1_buf[3*320*240];	// temp for downsized live image corrected for avg intensity
static guchar ref_buf[3*320*240];	// temp for reference picture buffer (time averaged)
static guchar res_buf[3*320*240];	// temp for motion analysis picture buffer (graphical results)
static guchar grf_buf[3*320*50];	// strip chart graph buffer 
struct video_picture v_pict;		// used to store prefered webcam palette

// the following is declared in main()... needed for lookup_widget() calls
extern GtkWidget *window1;

// the following is declared in fileio.c
extern char full_name[];        // a hack!. full path/filename of last written picture

// for setup items
GtkWidget *propertybox1;

gchar image_directory[1024];	// where the jpg's get put
gchar image_label[256];			// used by strftime to date stamp image
gchar video_device[256];		// opened to read images via v4l
gint picture_interval;			// seconds between pictures

float ref_update_fraction;		// amount a new image contributes to
								// reference image each cycle
gint alarm_threshold;			// percent of graph for alarm
float sig_pix_threshold;		// amount pixel must be different
								// (x*avg difference)		
float graph_scale;				// amount data is scaled before
								// being sent to 0-100 range plot																				
gint periodic_save_interval;	// force a save every xx snapshots

gint beep_on_alarm;				// console beep on alarm
gint show_target_box;           // puts active area on pictures
gchar alarm_command[1025];      // user command to execute on alarms


int dev;						// handle for v4l video camera device

struct alarm_zone alarm_zone;
unsigned int  	avg_ref_intensity;
unsigned int	avg_ds_intensity;
unsigned int	avg_ds1_intensity;
unsigned int	alarm_pix = 0;

/* The file selection widget and the string to store the chosen filename */
GtkWidget *file_selector;


gchar *selected_filename;

//=====================================================================
//=====================================================================
void on_exit1_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	printf(_("on_exit1_activate\n"));
	v4l_close(dev);	
	gtk_main_quit();	
}

//=====================================================================
//=====================================================================
void
on_preferences1_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  	GtkWidget *wid;
	char buf[1025];
	
  	propertybox1 = create_propertybox1 ();
	
	// grey out the help button.... none written yet
	gnome_dialog_set_sensitive((GnomeDialog*)propertybox1, 3, FALSE);	

	// set all the displays to whatever the current memory settings are
	wid = lookup_widget(GTK_WIDGET(propertybox1),"image_dir");
	gtk_entry_set_text(GTK_ENTRY(wid),image_directory);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"img_label");
	gtk_entry_set_text(GTK_ENTRY(wid),image_label);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"Pix_interval");
	sprintf(buf,"%d",picture_interval);
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"save_interval");
	sprintf(buf,"%d",periodic_save_interval);
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"video_src");
	gtk_entry_set_text(GTK_ENTRY(wid),video_device);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"ref_fraction");
	sprintf(buf,"%.2f",ref_update_fraction);
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"alrm_threshold");
	sprintf(buf,"%d",alarm_threshold);	
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"sig_pix_thres");
	sprintf(buf,"%.1f",sig_pix_threshold);	
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"plt_scaler");
	sprintf(buf,"%.2f",graph_scale);	
	gtk_entry_set_text(GTK_ENTRY(wid),buf);

	wid = lookup_widget(GTK_WIDGET(propertybox1),"beep_on_alrm");
	gtk_toggle_button_set_active((GtkToggleButton *)wid, beep_on_alarm);
    
	wid = lookup_widget(GTK_WIDGET(propertybox1),"show_target_box");
	gtk_toggle_button_set_active((GtkToggleButton *)wid, show_target_box);
    
	wid = lookup_widget(GTK_WIDGET(propertybox1),"alarm_command");
	sprintf(buf,"%s",alarm_command);	
	gtk_entry_set_text(GTK_ENTRY(wid),buf);
    
    
	
	// since setting all the boxes seems to mark items as changed,
	// set box box to unchanged so apply button is greyed out until
	// the user changes an item
	gnome_property_box_set_state((GnomePropertyBox *)propertybox1,FALSE);
		
  	gtk_widget_show (propertybox1);	
}


//=====================================================================
//=====================================================================
void
on_about1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  	GtkWidget *about2;

  	about2 = create_about2 ();	
  	gtk_widget_show (about2);	
}

//=====================================================================
// save image via menu file/save command, prefix pixname with 'SP'
// to keep it different from alarm pictures
//=====================================================================
void
on_save1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    time_t      t;
    struct tm  *tm;
    char        line[128];
		
	// write a copy to file because user asked for it
	time(&t);
    tm = localtime(&t);
    strftime(line,sizeof(line)-1,"SP%H%M%S.jpg",tm);	
	put_image_jpeg(line, rgb_buf, 640, 480, QUAL_DEFAULT);

}

//=====================================================================
// this routine gets called periodically (2 sec typically) and is the 
// driving force for capturing images.
//=====================================================================
gint timeout_callback( gpointer data )
{
	int size;
	char *image;
  	GtkWidget *wid;
	char tmp;
	char *img_ptr;
	int cnt;

	

	//grab a v4l frame
	if (dev <= 0)
	{
		g_print(_("Unable to use v4l camera device... open failed???"));	
		return TRUE;				// keep ticker happening
	}

	if(!v_pict.palette)
	{
		// get palette info
		if(ioctl(dev, VIDIOCGPICT, &v_pict)==-1)
		{
			g_print(_("Unable to get picture properties from webcam (VIDIOCGPICT failed)\n"));
			return TRUE;
		}
		if(v_pict.palette!=VIDEO_PALETTE_YUV420P && v_pict.palette!=VIDEO_PALETTE_RGB24)
		{
			// try to switch to RGB24 palette
			v_pict.palette=VIDEO_PALETTE_RGB24;
			if(ioctl(dev, VIDIOCSPICT, &v_pict)==-1)
			{
				// try to switch to YUV420P palette
				g_print(_("RGB24 not supported, trying YUV420P...\n"));
				v_pict.palette=VIDEO_PALETTE_YUV420P;
				if(ioctl(dev, VIDIOCSPICT, &v_pict)==-1)
				{
					g_print(_("YUV420P not supported. Unable to use this device\n"));
					return TRUE;
				}
			}
			
		}
	}
	
	image = get_image(dev,640,480,IN_DEFAULT,NORM_NTSC,v_pict.palette,&size);

	if ( image )
	{
		if(v_pict.palette==VIDEO_PALETTE_YUV420P)
		//YUV420P palette format
			ccvt_420p_rgb24(640,480,image,image+(640*480),image+(640*600),rgb_buf);
		else
		//BGR24 palette format
		{
			// grab a static memory copy for use in calcs and redraws
			memcpy(rgb_buf,image,sizeof(rgb_buf));
			// v4l returns bgr not rgb, so we need to swap red and blue
			// for our display
			img_ptr = rgb_buf;		
			for (cnt=0; cnt < 640*480; cnt++)
			{
				tmp = img_ptr[0];			// get blue for safe keeping
				img_ptr[0] = img_ptr[2];	// move red into correct place
				img_ptr[2] = tmp;			// park blue in correct place
				img_ptr += 3;				// bump to next triplet
			}
		}
    }
	else
		g_print(_("image capture failed\n"));
	
	// then dont forget to release buffers
	if (size)
    	munmap(image, size);
	else 
	if (image)
    	free(image);

	// let the widget know that it needs to be updated
	wid = GTK_WIDGET(lookup_widget(window1,"drawingarea1"));
	gtk_widget_queue_draw(wid);
	
	down_sample_image();		// convert 640x480 pix to 320x240 in ds_buf

	do_motion_analysis();		// check difference between ds_buf and ref_buf
	
	update_reference_image();	// avg ds_buf into ref_buff
	
	// keep us ticking... return 0 to terminate	

	return TRUE;
};

//=====================================================================
// given a box containing the alarm window, modify pixels in main
// display window to show box. Note main display is displayed at twice
// the x and y dimensions so scaling of the alarm_zone is required.
//=====================================================================
void mark_alarm_zone( struct alarm_zone *zone )
{
	int x,y;
	char *p1,*p2;

	if (zone->x1 < 0 || zone->x2 < 0 || zone->y1 < 0 || zone->y2 < 0 )
		return;
	if (zone->x1 > 639 || zone->x2 > 639 || zone->y1 > 479 || zone->y2 > 479 )
		return;
		
	p1 = rgb_buf + (zone->y1 * 2)*640*3;
	p2 = rgb_buf + (zone->y2 * 2)*640*3;
	for( x = 0; x < 640; x++)
	{
		*p1 = ~*p1;			// invert red component
		*(p1+1) = ~*(p1+1);	// invert green
		*(p1+2) = ~*(p1+2);	// invert blue

		*p2 = ~*p2;			// draw both horizontal lines
		*(p2+1) = ~*(p2+1);
		*(p2+2) = ~*(p2+2);	

		p1 += 3;
		p2 += 3;
	}

	p1 = rgb_buf + (zone->x1 * 2)*3;
	p2 = rgb_buf + (zone->x2 * 2)*3;	
	for( y = 0; y < 480; y++)
	{
		*p1 = ~*p1;			// invert red component
		*(p1+1) = ~*(p1+1);	// invert green
		*(p1+2) = ~*(p1+2);	// invert blue

		*p2 = ~*p2;			// draw both vertical lines
		*(p2+1) = ~*(p2+1);
		*(p2+2) = ~*(p2+2);	
		
		p1 += 640*3;
		p2 += 640*3;	
	}
}




//=====================================================================
// convert 640x480 picture down to 320x240
// use pixel averaging for best quality instead of subsampling
// 2 output buffers are filled...
// ds_buf[] has the downsampled image
// ds1_buf[] has a copy of the downsampled image corrected for
// 	intensity to make its avg intensity match the reference image
//=====================================================================
void down_sample_image(void)
{
	gint x,y;
	guchar *p0;
	guchar *p1;
	guchar *p2;
	guchar *p3;
	guchar *p4;
	guchar *pdest;
	unsigned int avg = 0;
	guchar correction;
	
	pdest = &ds_buf[0];
	for (y=0; y< 240; y++)
	{
		p1 = rgb_buf + (y*2) * 640 * 3;		// upper left cell
		p2 = p1 + 3;			            // upper right cell
		p3 = p1 + (640 * 3);				// lower left cell
		p4 = p3 + 3;						// lower right cell
		for (x=0; x<320; x++)
		{
			*pdest++ = (*p1++ + *p2++ + *p3++ + *p4++) / 4;	//red cells avg
			*pdest++ = (*p1++ + *p2++ + *p3++ + *p4++) / 4;	//grn cells avg
			*pdest++ = (*p1++ + *p2++ + *p3++ + *p4++) / 4;	//blu cells avg
			p1 += 3;
			p2 += 3;
			p3 += 3;
			p4 += 3;
		}
	}
	
	for ( p1 = &ds_buf[0]; p1 < &ds_buf[0]+sizeof(ds_buf); p1++ )
	{
		avg += *p1;
	}
	avg /= sizeof(ds_buf);
	avg_ds_intensity = avg;
//	g_print(_("downsampled image avg intensity = %u\n"),avg);

	// now generate an intensity normalized image
	p0 = &ds_buf[0];
	p1 = &ds1_buf[0];
	if (avg_ds_intensity < avg_ref_intensity)
	{
		correction = avg_ref_intensity - avg_ds_intensity;
//		g_print("making image lighter by %d\n",correction);		
		// make ds1 image brighter to match ref image
		for ( ; p1 < &ds1_buf[0]+sizeof(ds1_buf);p1++,p0++ )
		{
			*p1 = *p0 + correction;
			// fix up overflows
			if ( *p1 < *p0 ) *p1 = 255;
		}
	}
	else if(avg_ds_intensity == avg_ref_intensity)
	{
		// make exact copy
//		g_print("no image correction required\n");		
		memcpy(p1,p0,sizeof(ds1_buf));
	}
	
	else if(avg_ds_intensity > avg_ref_intensity)
	{
		correction = avg_ds_intensity - avg_ref_intensity;
//		g_print("making image darker by %d\n",correction);
		// make ds1 image darker to match ref image
		for (; p1 < &ds1_buf[0]+sizeof(ds1_buf); p1++,p0++ )
		{
			*p1 = *p0 - correction;
			// fix up underflows
			if ( *p1 > *p0 ) *p1 = 0;
		}
	
	}
	
	// for debug, its nice to know the avg intensity of the corrected buffer
	// it should be close to the ref buffer unless clipping occured due to
	// saturation at 0 or 255 limits
	for ( p1 = &ds1_buf[0]; p1 < &ds1_buf[0]+sizeof(ds1_buf); p1++ )
	{
		avg += *p1;
	}
	avg /= sizeof(ds1_buf);
	avg_ds1_intensity = avg;

}


//=====================================================================
// This procedure adds some of the new image in with lots of the old 
// image. This gives us a long time constant average reference picture.
// This allows the system to adapt to new items in the image over time
// but still alarm again if they move in the picture. Also, changing
// lighting conditions are less of a problem.
//=====================================================================
void update_reference_image(void)
{
  	GtkWidget *wid;
	gint x,y;
	guchar *pdest;
	guchar *psrc;
	float old;
	unsigned int avg = 0;
	guchar *p1;
	static int first = 1;
	
	if ( first == 1 )
	{
		// on first pass, the first picture becomes the reference picture
		memcpy(ref_buf,ds_buf,sizeof(ref_buf));
		first = 0;
	}
	else
	{		
		// now update reference buffer by using some of old image and some of new
		psrc  = &ds_buf[0];
		pdest = &ref_buf[0];	
		old = 1.0 - ref_update_fraction;
		for (y=0; y<240; y++)
		{
			for (x=0; x<320; x++)
			{
				// give the fpu something to do...
				*pdest++ = old * *pdest + ref_update_fraction * *psrc++ + 0.5;	// red
				*pdest++ = old * *pdest + ref_update_fraction * *psrc++ + 0.5;	// green
				*pdest++ = old * *pdest + ref_update_fraction * *psrc++ + 0.5;	// blue
			}
		}
	}

	// compute new avg intensity for the reference buffer
	for ( p1 = &ref_buf[0]; p1 < &ref_buf[0]+sizeof(ref_buf); p1++ )
	{
		avg += *p1;
	}
	avg /= sizeof(ref_buf);
//	g_print(_("reference image avg intensity = %u\n"),avg);
	avg_ref_intensity = (guchar)avg;
	
	// let the widget know that it needs to be updated
	wid = GTK_WIDGET(lookup_widget(window1,"drawingarea2"));
	gtk_widget_queue_draw(wid);
		
}
//=====================================================================
// call with a number 0..100 for graph update
//=====================================================================
void update_history_graph(int percent, int alarm_level)
{
  	GtkWidget *wid;
	static int first = 1;
	int		y;
	guchar *ptr;
	static int call_cnt = 0;	// used for threshold line on graph
	
	call_cnt++;
	
	// if first time through, clear graph to black
	if (first)
	{
		memset(&grf_buf[0],0,sizeof(grf_buf));
		first = 0;
	}
	
  	for (y=0; y<50; y++)
  	{
	  	// move each row of pixels over in memory (scroll left)
	  	memmove( grf_buf+y*320*3,grf_buf+y*320*3+3,319*3); 

	  	// init last column to black
	  	ptr = grf_buf + y*320*3 + 319*3;
	  	*ptr++ = 0;  	  //red
	  	*ptr++ = 0;  	  //grn
	  	*ptr++ = 0;  	  //blu

  	}
	
	// turn on 1 pixel in last column
	// limit range so pixel always gets displayed 0...99
	if (percent < 0)
		percent = 0;
	if (percent > 99)
		percent = 99;

	// invert graph so that 0 is at bottom, scale range to fit 50 pixel image
	ptr = grf_buf + ((99-percent)/2)*320*3 + 319*3;
	while ( ptr < grf_buf + sizeof(grf_buf))
	{
		if ( percent > alarm_level )
		{		
			*ptr = 255;
			*(ptr+1) = 32;
			*(ptr+2) = 32;
		}
		else
		{
			*ptr = 32;
			*(ptr+1) = 255;
			*(ptr+2) = 32;
		}
		ptr += 320 * 3;	// bump down 1 line
	}		

	// time to put a reference dot on the plot showing alarm threshold
	if ( call_cnt %4 == 0)
	{
		if ( alarm_level < 0)
			alarm_level = 0;
		if (alarm_level > 99)
			alarm_level = 99;
		ptr = grf_buf + ((99-alarm_level)/2)*320*3 + 319*3;
		*ptr++ = 32;
		*ptr++ = 255;
		*ptr = 255;
	}

	// let the widget know that it needs to be updated
	wid = GTK_WIDGET(lookup_widget(window1,"drawingarea4"));
	gtk_widget_queue_draw(wid);
}


//=====================================================================
// routine to subtract downsized image from the reference image
// this will give us an image with only the changes showing
// its not quite a simple subtraction, we look at each color and
// look for color changes as well. Significantly changed pixels are
// set to bright white. Result is a b/w image with additional
// blob filtering to look for areas of change (marked as red pixels).
// Each blob is a pixel surrounded by 8 neighbors
//=====================================================================
float do_motion_analysis(void)
{
  	GtkWidget *wid;
	gint x,y;
	guchar *pdest;
	guchar *psrc;
	guchar *pref;
	guchar red;
	guchar green;
	guchar blue;
	guchar cresult;
	float  fresult = 0.0f;
	int		pixcnt;
	guchar threshold;
	int    blob_cnt;
	int i;	
    time_t      t;
    struct tm  *tm;
    char        line[128];
	static	int image_cnt = 0;		// used to force saves on periodic basis
	guint context;
	char sbuf[256];
	static int first = 1;
    char *cp;
    gchar **toks;
    gchar *new_cmd;
        
    	
	// on the first pass, we dont have a ref image stored yet so
	// we cannot do a difference that is meaningful.
	if ( first == 1 )
	{
		first = 0;
		return 0.0;
	}
	
	alarm_zone.x1 = 640;			// init bounding box of alarm blobs
	alarm_zone.x2 = -1;
	alarm_zone.y1 = 480;
	alarm_zone.y2 = -1;

//	g_print(_("new image is %s, %u %u %u\n"),
//			( avg_ds_intensity <= avg_ref_intensity ) ? _("darker or ==") : _("lighter"),
//			avg_ds_intensity, avg_ds1_intensity, avg_ref_intensity);	
	pref  = &ref_buf[0];		
	psrc  = &ds1_buf[0];	// intensity corrected buffer
	pdest = &res_buf[0];
	
	for (y=0; y<240; y++)
	{
		for (x=0; x<320; x++)
		{
		
			// give the cpu something to do...
			// compute rms diff between all pixels
			// detection of color or intensity changes detected
			red = (*pref > *psrc) ? *pref- *psrc : *psrc- *pref;	// red
			pref++; psrc++;
			green = (*pref > *psrc) ? *pref- *psrc : *psrc- *pref;;	// green
			pref++; psrc++;
			blue  = (*pref > *psrc) ? *pref- *psrc : *psrc- *pref;;	// blue
			pref++; psrc++;
			cresult = sqrt((red*red + green*green + blue*blue)/3);
			*pdest++ = cresult;		// make b/w image
			*pdest++ = cresult;
			*pdest++ = cresult;
			fresult += (float)cresult;
		}
	}
	// calculate normalized difference over picture
	fresult = fresult/(float)(320*240);
	
	// detect and mark pixels that are significantly different
	threshold = (guchar)sig_pix_threshold;
	pixcnt = 0;
	psrc  = &res_buf[0];
	while ( psrc < res_buf+sizeof(res_buf))
	{
		if ( *psrc > threshold )
		{
			pixcnt++;
			*psrc = 255;		// alarm active areas in white
			*(psrc+1) = 255;
			*(psrc+2) = 255;
		}
			
		psrc += 3;			// only need to look at reds of each pixel
							// because g and b are set to the same thing			
	}
	
	// to eliminate small changes (wind blown objects) from causing so much
	// havoc, we will try to eliminate false alarms by scanning for blobs	
	// A blob is a 3x3 pixel area, all must be 'hot' (white)
	blob_cnt = 0;
	psrc  = res_buf + 320 * 3;
	while ( psrc < (res_buf+sizeof(res_buf)-320*3))
	{
		for (i=0;i<1;i++)
		{
			if (*psrc < 255) break;			// central pixel must be lite
			if (*(psrc-320*3) < 255) break;	// pixel above must be lite
			if (*(psrc+320*3) < 255) break;	// pixel below must be lite
			if (*(psrc-3) < 255) break;		// pixel to left must be lite
			if (*(psrc+3) < 255) break;		// pixel to right must be light
			if (*(psrc-320*3+3) < 255) break;		// pixel to  NE must be light			
			if (*(psrc-320*3-3) < 255) break;		// pixel to  NW must be light						
			if (*(psrc+320*3+3) < 255) break;		// pixel to  SE must be light						
			if (*(psrc+320*3-3) < 255) break;		// pixel to  SW must be light						
			// mark blob as blob on screen
			*(psrc+2) = 0;	// shut off blue
			*(psrc+1) = 0;  // shut off green
			blob_cnt++;
			x = ((psrc - res_buf) % (320 * 3))/3;
			y = (psrc - res_buf) / (320 * 3);
			if ( x < alarm_zone.x1 )		// set first occurance location
				alarm_zone.x1 = x;
			if ( alarm_zone.x2 < x )		// keep highest occurance
				alarm_zone.x2 = x;			
			if ( y < alarm_zone.y1 )		// set first occurance location
				alarm_zone.y1 = y;
			if ( alarm_zone.y2 < y )		// keep highest occurance
				alarm_zone.y2 = y;			
		}
		psrc += 3;
	}

	// update status bar with some of our calcs
	wid = lookup_widget(window1,"statusbar1");
	context = gtk_statusbar_get_context_id(GTK_STATUSBAR(wid),"do_motion_analysis()");
	sprintf(sbuf,_("Image%d : alarmpix = %u   alarm pixels = %d, alarm blobs = %d rmsdiff=%.1f"),
    	    image_cnt,alarm_pix,pixcnt,blob_cnt, fresult);
	gtk_statusbar_pop(GTK_STATUSBAR(wid),context);
	gtk_statusbar_push(GTK_STATUSBAR(wid),context,sbuf);
	
	// let the widget know that it needs to be updated
	wid = GTK_WIDGET(lookup_widget(window1,"drawingarea3"));
	gtk_widget_queue_draw(wid);

	// label main view with time/date location stamp (display and filesave)
	if (blob_cnt*graph_scale > alarm_threshold)
	{
		alarm_pix++;
        if ( show_target_box )
		    mark_alarm_zone(&alarm_zone);			// box in alarm window
		add_rgb_text(rgb_buf,640,480, 1);		// red text lable on pix
		if (beep_on_alarm)						// wakeup the operator
			gdk_beep();
	}
	else
		add_rgb_text(rgb_buf,640,480, 0);		// white text label on pix	
		
	
	image_cnt++;	// save image on motion or every xxx pictures

	if (( blob_cnt*graph_scale > alarm_threshold ) || 
		((image_cnt % periodic_save_interval) == 0))
	{
    	time(&t);
    	tm = localtime(&t);
    	strftime(line,sizeof(line)-1,"GSPY%H%M%S.jpg",tm);	
		put_image_jpeg(line, rgb_buf, 640, 480, 80);
	}
    
    // on each alarm, execute a user command
    if (( blob_cnt*graph_scale > alarm_threshold ) &&
        ( strlen(alarm_command) > 1 ) )
    {
        if ((cp = strstr(alarm_command,"%f%")) == NULL)
            // no filename substitution required in user provided string
            system(alarm_command);
        else
        {
            // user put filename(s) in command line using %f%, then we need
            // to insert the current filename into the user command string

            toks = g_strsplit(alarm_command,"%f%",0);
            new_cmd = g_strjoinv(full_name,toks);
            system(new_cmd);
            free(new_cmd);
            g_strfreev(toks);
            

        }
    }    
	update_history_graph(blob_cnt*graph_scale,alarm_threshold);	

	return fresult;
}
//=====================================================================
//=====================================================================
gboolean on_drawingarea1_expose_event  (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	gdk_draw_rgb_image(widget->window,widget->style->fg_gc[GTK_STATE_NORMAL],
	           0,0,640,480,GDK_RGB_DITHER_NORMAL,rgb_buf,3*640); 
 	return FALSE;
}

//=====================================================================
//=====================================================================
gboolean on_drawingarea2_expose_event  (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	gdk_draw_rgb_image(widget->window,widget->style->fg_gc[GTK_STATE_NORMAL],
	           0,0,320,240,GDK_RGB_DITHER_NORMAL,ref_buf,3*320); 
  	return FALSE;
}

//=====================================================================
//=====================================================================
gboolean on_drawingarea3_expose_event  (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	gdk_draw_rgb_image(widget->window,widget->style->fg_gc[GTK_STATE_NORMAL],
	           0,0,320,240,GDK_RGB_DITHER_NORMAL,res_buf,3*320); 
  	return FALSE;
}

//=====================================================================
//=====================================================================
gboolean on_drawingarea4_expose_event  (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
	gdk_draw_rgb_image(widget->window,widget->style->fg_gc[GTK_STATE_NORMAL],
	           0,0,320,50,GDK_RGB_DITHER_NORMAL,grf_buf,3*320); 

  return FALSE;
}


//=====================================================================
// we get here whenever a user has changed a setup item
// we need to activate the apply button so user knows items have changed
// all data entry items use this as the "changed" event handler
//=====================================================================
void
apply_needed                           (GtkEditable     *editable,
                                        gpointer         user_data)
{
	gnome_property_box_changed((GnomePropertyBox*) propertybox1); 		
}

//=====================================================================
// we get here when the user has clicked apply or the ok button
// in the setup windows
//=====================================================================
void
on_propertybox1_apply                  (GnomePropertyBox *gnomepropertybox,
                                        gint             arg1,
                                        gpointer         user_data)
{
  	GtkWidget *wid;
	gchar *buf;
	// arg refers to the page that has changed
	// with a final value of -1... normally, just save everything on -1
	
	if ( arg1 == -1 )
	{
		//"save all setup items here to ~/.gnome/gspy (like ini file)
		
		wid = lookup_widget(GTK_WIDGET(propertybox1),"image_dir");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		strncpy(image_directory,buf,sizeof(image_directory));
		gnome_config_set_string("/gspy/general/image_directory", image_directory);	// file, section, key
		
		wid = lookup_widget(GTK_WIDGET(propertybox1),"img_label");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		strncpy(image_label,buf,sizeof(image_label));
		gnome_config_set_string("/gspy/general/image_label", image_label);	// file, section, key
			
		wid = lookup_widget(GTK_WIDGET(propertybox1),"Pix_interval");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		picture_interval = atoi(buf);
		gnome_config_set_int("/gspy/general/picture_interval", picture_interval);	// file, section, key	

		wid = lookup_widget(GTK_WIDGET(propertybox1),"save_interval");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		periodic_save_interval = atoi(buf);
		gnome_config_set_int("/gspy/general/periodic_save_interval", periodic_save_interval);	// file, section, key	

		wid = lookup_widget(GTK_WIDGET(propertybox1),"video_src");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		strncpy(video_device,buf,sizeof(video_device));	
		gnome_config_set_string("/gspy/video4linux/video_device", video_device);	// file, section, key
	
		wid = lookup_widget(GTK_WIDGET(propertybox1),"ref_fraction");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));	
		ref_update_fraction = atof(buf);
		gnome_config_set_float("/gspy/alarms/ref_update_fraction",ref_update_fraction );	// file, section, key

		wid = lookup_widget(GTK_WIDGET(propertybox1),"alrm_threshold");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		alarm_threshold = atoi( buf );
		gnome_config_set_int("/gspy/alarms/alarm_threshold",alarm_threshold );	// file, section, key	
			
		wid = lookup_widget(GTK_WIDGET(propertybox1),"sig_pix_thres");
		buf= gtk_entry_get_text(GTK_ENTRY(wid));
		sig_pix_threshold = atof( buf );
		gnome_config_set_float("/gspy/alarms/sig_pix_threshold", sig_pix_threshold);	// file, section, key	

		wid = lookup_widget(GTK_WIDGET(propertybox1),"plt_scaler");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		graph_scale = atof( buf );
		gnome_config_set_float("/gspy/alarms/graph_scale", graph_scale);	// file, section, key		

		wid = lookup_widget(GTK_WIDGET(gnomepropertybox),"beep_on_alrm");
		beep_on_alarm = (GTK_TOGGLE_BUTTON (wid)->active);
		gnome_config_set_bool("/gspy/alarms/beep_on_alarm",beep_on_alarm );	// file, section, key
        
        wid = lookup_widget(GTK_WIDGET(gnomepropertybox),"show_target_box");
		show_target_box = (GTK_TOGGLE_BUTTON (wid)->active);
		gnome_config_set_bool("/gspy/alarms/show_target_box",show_target_box );	// file, section, key

		wid = lookup_widget(GTK_WIDGET(propertybox1),"alarm_command");
		buf = gtk_entry_get_text(GTK_ENTRY(wid));
		strncpy(alarm_command,buf,sizeof(alarm_command));

        // filename substitution works better with an extra "token"
        if(strlen(alarm_command) < (sizeof(alarm_command)-2))
            strcat(alarm_command," ");
            	
		gnome_config_set_string("/gspy/alarms/alarm_command", alarm_command);	// file, section, key
	
        
		// force out stuff to disk .. yea I'm paranoid
		gnome_config_sync();	
	
	}
}

//=====================================================================
// restore all the non volatile preferences.. if they have never been
// saved or the config file got deleted, program should be close to 
// on the default compile values.
//=====================================================================
void restore_setup_preferences( void )
{
	gchar 	*var;
	gint 	def;
	
	if (!gnome_config_has_section("/gspy/general/"))
	{		
 		GtkWidget * mbox;
   		mbox = gnome_message_box_new(
			_("Setup File Not Found... \n"
              "A default setup file will be created.\n"
              "Please check setup preferences"), 
            GNOME_MESSAGE_BOX_INFO,
            GNOME_STOCK_BUTTON_OK, 
            NULL);
   		gtk_widget_show (mbox);
	}

	var = gnome_config_get_string_with_default(
	       "/gspy/general/image_directory=./",&def);
	strncpy(image_directory,var,sizeof(image_directory));
	g_free(var);


	var = gnome_config_get_string_with_default(
	      	"/gspy/general/image_label=Room 204 %Y-%m-%d %H:%M:%S",&def);
	strncpy(image_label,var,sizeof(image_label));
	g_free(var);	

	picture_interval = gnome_config_get_int_with_default(
		"/gspy/general/picture_interval=2",&def);

	periodic_save_interval = gnome_config_get_int_with_default(
		"/gspy/general/periodic_save_interval=300",&def);

	var = gnome_config_get_string_with_default(
		"/gspy/video4linux/video_device=/dev/video0",&def);
	strncpy(video_device,var,sizeof(video_device));		
	g_free(var);

	ref_update_fraction = gnome_config_get_float_with_default(
		"/gspy/alarms/ref_update_fraction=0.2",&def);
		
	alarm_threshold = gnome_config_get_int_with_default(
		"/gspy/alarms/alarm_threshold=70",&def);

	sig_pix_threshold = gnome_config_get_float_with_default(
		"/gspy/alarms/sig_pix_threshold=25.0",&def);

	graph_scale = gnome_config_get_float_with_default(
		"/gspy/alarms/graph_scale=1.0",&def);

	beep_on_alarm = gnome_config_get_bool_with_default(
		"/gspy/alarms/beep_on_alarm=1",&def);
        
    show_target_box = gnome_config_get_bool_with_default(
		"/gspy/alarms/show_target_box=1",&def);	

	var = gnome_config_get_string_with_default(
		"/gspy/alarms/alarm_command= ls -l %f% &",&def);
	strncpy(alarm_command,var,sizeof(alarm_command));		
	g_free(var);


}



//======================================================================
//	Signal Handler for SIGCHILD signals used when forking to generate 
//  mpg files.
//======================================================================
void sigchld_handler(int num)
{
	sigset_t set, oldset;
	pid_t pid;
	int status, exitstatus;

	/* block other incoming SIGCHLD signals */
	sigemptyset(&set);
	sigaddset(&set, SIGCHLD);
	sigprocmask(SIG_BLOCK, &set, &oldset);

	/* wait for child */
	while((pid = waitpid((pid_t)-1, &status, WNOHANG)) > 0)
  	{
		if(WIFEXITED(status))
	  	{
			exitstatus = WEXITSTATUS(status);

			fprintf(stderr, 
				"Parent: child exited, pid = %d, exit status = %d\n", 
				(int)pid, exitstatus);
	  	}
		else if(WIFSIGNALED(status))
	  	{
			exitstatus = WTERMSIG(status);

			fprintf(stderr,
				"Parent: child terminated by signal %d, pid = %d\n",
				exitstatus, (int)pid);
	  	}
		else if(WIFSTOPPED(status))
	  	{
			exitstatus = WSTOPSIG(status);

			fprintf(stderr,
				"Parent: child stopped by signal %d, pid = %d\n",
				exitstatus, (int)pid);
	  	}
		else
	  	{
			fprintf(stderr,
				"Parent: child exited magically, pid = %d\n",
				(int)pid);
	  	}
	}

	/* re-install the signal handler (some systems need this) */
	signal(SIGCHLD, sigchld_handler);

	/* and unblock it */
	sigemptyset(&set);
	sigaddset(&set, SIGCHLD);
	sigprocmask(SIG_UNBLOCK, &set, &oldset);
}



// helper function for selecting files from directory
static int one(const struct dirent *entry)
{
	if (strstr(entry->d_name,".jpg"))
		return 1;
	else
		return 0;
}

void make_mpeg_part2(GtkFileSelection *selector, gpointer user_data) 
{
  	pid_t pid;
    struct dirent **eps;
    int n;

    selected_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION(file_selector));
	g_print(_("user selected directory %s\n"),selected_filename);

  	pid = fork();

  	if(pid == -1)
	{
	  	/* ouch, fork() failed */
	  	perror("fork");
	  	exit(-1);
	}
  	else if(pid == 0)
	{
		FILE *fp=NULL;
		
	  	/* child */
	  	fprintf(stderr, "Child: pid = %d\n", (int)getpid());
		// we need to build the config file for the mpeg encoder
		// and then call the mpeg encoder
       	n = scandir(selected_filename, &eps, one, alphasort);
		if (n == 0)
		{
			g_print(_("Child: No jpg files found in directory selected\n"));
			_exit(-1);
		}
       	if (n > 0)
        {
			int cnt;
			g_print(_("Child: found %d files \n"),n);
			fp=fopen("video.params", "w");
			if (!fp) 
			{	
				g_print(_("Child: Failed to open/create video.params file\n"));
				_exit(-1);
			}
			fprintf(fp,
			"PATTERN IBBPBBPBBPBBPBBP\n"
			"OUTPUT %svideo.mpg\n"
			"BASE_FILE_FORMAT JPEG\n"
			"INPUT_CONVERT *\n"
			"GOP_SIZE 30\n"
			"SLICES_PER_FRAME 1\n"
			"INPUT_DIR %s\n"
			"INPUT\n",selected_filename,selected_filename);
			for (cnt = 0; cnt < n; ++cnt)
				fprintf(fp,"%s\n",eps[cnt]->d_name);
				
			fprintf(fp,
				"END_INPUT\n"
				"PIXEL HALF\n"
				"RANGE 10\n"
				"PSEARCH_ALG LOGARITHMIC\n"
				"BSEARCH_ALG CROSS2\n"
				"IQSCALE 8\n"
				"PQSCALE 10\n"
				"BQSCALE 25\n"
				"FORCE_ENCODE_LAST_FRAME\n"
				"REFERENCE_FRAME ORIGINAL\n"
				"FRAME_RATE 25\n");
			fclose(fp);
			nice(5);
			g_print("Child: executing mpeg encoder...\n");
			execlp("mpeg_encode", "mpeg_encode", "video.params", 0);
	  		/* if exec() returns, there is something wrong */
			perror("Child: Failed to start mpeg_encode");
        }
       	else
         	perror(_("Child: Couldn't open the directory"));


	  	/* exit child. note the use of _exit() instead of exit() */
	  	_exit(-1);
	}
  	else
	{
	  	/* parent */
	  	fprintf(stderr, "Parent: forked a child with pid = %d\n", (int)pid);
	}
}


void on_makempeg1_activate(GtkMenuItem *menuitem, gpointer user_data)
{
	g_print(_("make me an mpeg\n"));
   /* Create the selector */
    
    file_selector = gtk_file_selection_new(_("Please select a directory of .jpg files"));
    
    gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
                            "clicked", GTK_SIGNAL_FUNC (make_mpeg_part2), NULL);
                            
    /* Ensure that the dialog box is destroyed when the user clicks a button. */
    
    gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
                                           "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
                                           (gpointer) file_selector);

    gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->cancel_button),
                                           "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
                                           (gpointer) file_selector);

	gtk_file_selection_set_filename((GtkFileSelection *)file_selector,(const gchar *)image_directory);    
    /* Display that dialog */   
    gtk_widget_show (file_selector);	
}


